from itertools import product

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

import astropy
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from astropy.units import Quantity
from astropy.wcs.wcsapi import HighLevelWCSWrapper

from ndcube.wcs.wrappers import CompoundLowLevelWCS


@pytest.fixture
def spectral_wcs(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def celestial_wcs(request):
    return request.getfixturevalue(request.param)


EXPECTED_CELESTIAL_SPECTRAL_APE14_REPR = """
CompoundLowLevelWCS Transformation

This transformation has 3 pixel and 3 world dimensions

Array shape (Numpy order): (7, 6, 3)

Pixel Dim  Axis Name  Data size  Bounds
        0  None               3  (1, 3)
        1  None               6  (-1, 5)
        2  None               7  (1, 7)

World Dim  Axis Name        Physical Type  Units
        0  Frequency        em.freq        Hz
        1  Right Ascension  pos.eq.ra      deg
        2  Declination      pos.eq.dec     deg

Correlation between pixel and world axes:

             Pixel Dim
World Dim    0    1    2
        0  yes   no   no
        1   no  yes  yes
        2   no  yes  yes
""".strip()


@pytest.mark.parametrize(('spectral_wcs', 'celestial_wcs'),
                         product(['spectral_1d_ape14_wcs', 'spectral_1d_fitswcs'],
                                 ['celestial_2d_ape14_wcs', 'celestial_2d_fitswcs']),
                         indirect=True)
def test_celestial_spectral_ape14(spectral_wcs, celestial_wcs):

    wcs = CompoundLowLevelWCS(spectral_wcs, celestial_wcs)

    assert wcs.pixel_n_dim == 3
    assert wcs.world_n_dim == 3
    assert tuple(wcs.world_axis_physical_types) == ('em.freq', 'pos.eq.ra', 'pos.eq.dec')
    assert tuple(wcs.world_axis_units) == ('Hz', 'deg', 'deg')
    assert tuple(wcs.pixel_axis_names) == ('', '', '')
    assert tuple(wcs.world_axis_names) == ('Frequency',
                                           'Right Ascension',
                                           'Declination')
    assert_equal(wcs.axis_correlation_matrix, np.array([[1, 0, 0],
                                                        [0, 1, 1],
                                                        [0, 1, 1]]))

    # If any of the individual shapes are None, return None overall
    assert wcs.pixel_shape is None
    assert wcs.array_shape is None
    assert wcs.pixel_bounds == ((np.iinfo(int).min, np.iinfo(int).max), (-1, 5), (1, 7))

    # Set the shape and bounds on the spectrum and test again
    spectral_wcs.pixel_shape = (3,)
    spectral_wcs.pixel_bounds = [(1, 3)]
    assert wcs.pixel_shape == (3, 6, 7)
    assert wcs.array_shape == (7, 6, 3)
    assert wcs.pixel_bounds == ((1, 3), (-1, 5), (1, 7))

    pixel_scalar = (2.3, 4.3, 1.3)
    world_scalar = (-1.91e10, 5.4, -9.4)
    assert_allclose(wcs.pixel_to_world_values(*pixel_scalar), world_scalar)
    assert_allclose(wcs.array_index_to_world_values(*pixel_scalar[::-1]), world_scalar)
    assert_allclose(wcs.world_to_pixel_values(*world_scalar), pixel_scalar, equal_nan=True)

    assert str(wcs) == EXPECTED_CELESTIAL_SPECTRAL_APE14_REPR
    assert EXPECTED_CELESTIAL_SPECTRAL_APE14_REPR in repr(wcs)

    # Not going to test too many things with out of bounds inputs
    spectral_wcs.pixel_bounds = None
    celestial_wcs.pixel_bounds = None
    assert wcs.pixel_bounds is None
    assert_allclose(wcs.world_to_array_index_values(*world_scalar), [1, 4, 2])

    pixel_array = (np.array([2.3, 2.4]),
                   np.array([4.3, 4.4]),
                   np.array([1.3, 1.4]))
    world_array = (np.array([-1.91e10, -1.88e10]),
                   np.array([5.4, 5.2]),
                   np.array([-9.4, -9.2]))
    assert_allclose(wcs.pixel_to_world_values(*pixel_array), world_array)
    assert_allclose(wcs.array_index_to_world_values(*pixel_array[::-1]), world_array)
    assert_allclose(wcs.world_to_pixel_values(*world_array), pixel_array)
    assert_allclose(wcs.world_to_array_index_values(*world_array),
                    [[1, 1], [4, 4], [2, 2]])

    wcs_hl = HighLevelWCSWrapper(wcs)

    spectral, celestial = wcs_hl.pixel_to_world(*pixel_scalar)
    assert isinstance(spectral, Quantity)
    assert_quantity_allclose(spectral, world_scalar[0] * u.Hz)
    assert isinstance(celestial, SkyCoord)
    assert_quantity_allclose(celestial.ra, world_scalar[1] * u.deg)
    assert_quantity_allclose(celestial.dec, world_scalar[2] * u.deg)

    spectral, celestial = wcs_hl.pixel_to_world(*pixel_array)
    assert isinstance(spectral, Quantity)
    assert_quantity_allclose(spectral, world_array[0] * u.Hz)
    assert isinstance(celestial, SkyCoord)
    assert_quantity_allclose(celestial.ra, world_array[1] * u.deg)
    assert_quantity_allclose(celestial.dec, world_array[2] * u.deg)


def test_shared_pixel_axis_compound_1d(spectral_1d_fitswcs, time_1d_fitswcs):

    wcs = CompoundLowLevelWCS(spectral_1d_fitswcs, time_1d_fitswcs, mapping=(0, 0))

    assert wcs.world_axis_names == ("Frequency", "Time")
    assert wcs.world_axis_physical_types == ("em.freq", "time")
    assert wcs.world_axis_units == ("Hz", "s")
    assert wcs.world_n_dim == 2

    assert wcs.pixel_n_dim == 1
    assert wcs.pixel_shape is None
    assert wcs.pixel_axis_names == ('',)
    assert wcs.pixel_bounds is None

    world = wcs.pixel_to_world_values(0)
    np.testing.assert_allclose(world, (-2.6e+10, -7.))
    np.testing.assert_allclose(wcs.world_to_pixel_values(*world), 0)

    np.testing.assert_allclose(wcs.axis_correlation_matrix, [[True],
                                                             [True]])


def test_shared_pixel_axis_compound_3d(spectral_cube_3d_fitswcs, time_1d_fitswcs):

    spectral_cube_3d_fitswcs._naxis = [10, 20, 30]
    time_1d_fitswcs._naxis = [20]

    wcs = CompoundLowLevelWCS(spectral_cube_3d_fitswcs, time_1d_fitswcs, mapping=(0, 1, 2, 1))

    assert wcs.world_axis_names == ("Right Ascension", "Declination", "Frequency", "Time")
    assert wcs.world_axis_physical_types == ("pos.eq.ra", "pos.eq.dec", "em.freq", "time")
    assert wcs.world_axis_units == ("deg", "deg", "Hz", "s")
    assert wcs.world_n_dim == 4

    assert wcs.pixel_n_dim == 3
    np.testing.assert_allclose(wcs.pixel_shape, (10, 20, 30))
    assert wcs.pixel_axis_names == ('', '', '')
    assert wcs.pixel_bounds == ((-1, 5), (1, 7), (1, 2.5))

    np.testing.assert_allclose(wcs.axis_correlation_matrix, [[True, True, False],
                                                             [True, True, False],
                                                             [False, False, True],
                                                             [False, True, False]])

    world = wcs.pixel_to_world_values(0, 0, 0)
    if astropy.__version__ >= "7.0.0":
        np.testing.assert_allclose(world, (14, np.nan, np.nan, -7.0))
        np.testing.assert_allclose(wcs.world_to_pixel_values(*world), (0, np.nan, np.nan))
    else:
        np.testing.assert_allclose(world, (14, -12, -2.6e+10, -7.0))
        np.testing.assert_allclose(wcs.world_to_pixel_values(*world), (0, 0, 0))

    with pytest.raises(ValueError):
        wcs.world_to_pixel_values((14, -12, -2.6e+10, -6.0))

    with pytest.raises(ValueError):
        wcs.world_to_pixel_values((14, -10, -2.6e+10, -7.0))



import abc
from typing import Any
from numbers import Integral
from functools import reduce, partial
from collections.abc import Iterable

import numpy as np

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy.wcs import WCS
from astropy.wcs.wcsapi import BaseHighLevelWCS
from astropy.wcs.wcsapi.high_level_wcs_wrapper import HighLevelWCSWrapper
from astropy.wcs.wcsapi.wrappers.sliced_wcs import SlicedLowLevelWCS, sanitize_slices

from ndcube.utils.wcs import convert_between_array_and_pixel_axes
from ndcube.wcs.wrappers import CompoundLowLevelWCS, ResampledLowLevelWCS
from .table_coord import (
    BaseTableCoordinate,
    MultipleTableCoordinate,
    QuantityTableCoordinate,
    SkyCoordTableCoordinate,
    TimeTableCoordinate,
)

__all__ = ['ExtraCoords', 'ExtraCoordsABC']


class ExtraCoordsABC(abc.ABC):
    """
    A representation of additional world coordinates associated with pixel axes.

    ExtraCoords can be initialised by either specifying a
    `~astropy.wcs.wcsapi.BaseLowLevelWCS` object and a ``mapping``, or it can be
    built up by specifying one or more lookup tables.

    Parameters
    ----------
    wcs
        The WCS specifying the extra coordinates.
    mapping
       The mapping between the array dimensions and pixel dimensions in the
       extra coords object. This is an iterable of ``(array_dimension, pixel_dimension)`` pairs
       of length equal to the number of pixel dimensions in the extra coords.

    """
    @abc.abstractmethod
    def add(self,
            name: str | Iterable[str],
            array_dimension: int | Iterable[int],
            lookup_table: Any,
            physical_types: str | Iterable[str] = None,
            **kwargs):
        """
        Add a coordinate to this `~ndcube.ExtraCoords` based on a lookup table.

        Parameters
        ----------
        name : `str` or sequence of `str`
            The name(s) for these world coordinate(s).
        array_dimension : `int` or `tuple` of `int`
            The array dimension(s), to which this lookup table corresponds.
        lookup_table : `object` or sequence of `object`
            The lookup table. A `~ndcube.extra_coords.BaseTableCoordinate` subclass or anything
            that can instantiate one, i.e. currently a `~astropy.time.Time`,
            `~astropy.coordinates.SkyCoord`, or a (sequence of) `~astropy.units.Quantity`.
        physical_types: `str` or iterable of `str`, optional
            Descriptor(s) of the :ref:`<dimensions and physical types <dimensions>`
            associated with each axis; length must match the number of dimensions in
            ``lookup_table``.
        """

    @abc.abstractmethod
    def keys(self) -> Iterable[str]:
        """
        The world axis names for all the coordinates in the extra coords.
        """

    @property
    @abc.abstractmethod
    def mapping(self) -> Iterable[tuple[int, int]]:
        """
        The mapping between the array dimensions and pixel dimensions.

        This is an iterable of ``(array_dimension, pixel_dimension)`` pairs
        of length equal to the number of pixel dimensions in the extra coords.
        """

    @property
    @abc.abstractmethod
    def wcs(self) -> BaseHighLevelWCS:
        """
        A WCS object representing the world coordinates described by this ``ExtraCoords``.

        .. note::
            This WCS object does not map to the pixel dimensions of the data array
            in the `~ndcube.NDCube` object. It only includes pixel dimensions associated
            with the extra coordinates. For example, if there is only one extra coordinate
            associated with a single pixel dimension, this WCS will only have 1 pixel dimension,
            even if the `~ndcube.NDCube` object has a data array of 2-D or greater.
            Therefore using this WCS directly might lead to some confusing results.

        """

    @property
    @abc.abstractproperty
    def is_empty(self):
        """Return True if no extra coords present, else return False."""

    @abc.abstractmethod
    def __getitem__(self, item: str | int | slice | Iterable[str | int | slice]) -> "ExtraCoordsABC":
        """
        ExtraCoords can be sliced with either a string, or a numpy like slice.

        When sliced with a string it should return a new ExtraCoords object
        with only those coordinates with the given names. When sliced with a
        numpy array like slice it should return a new ExtraCoords with the
        slice applied. Supporting step is not required and "fancy indexing" is
        not supported.
        """


class ExtraCoords(ExtraCoordsABC):
    """
    A representation of additional world coordinates associated with pixel axes.

    ExtraCoords can be initialised by either specifying a
    `~astropy.wcs.wcsapi.BaseLowLevelWCS` object and a ``mapping``, or it can be
    built up by specifying one or more lookup tables.

    Parameters
    ----------
    wcs
        The WCS specifying the extra coordinates.
    mapping
       The mapping between the array dimensions and pixel dimensions in the
       extra coords object. This is an iterable of ``(array_dimension, pixel_dimension)`` pairs
       of length equal to the number of pixel dimensions in the extra coords.

    """

    def __init__(self, ndcube=None):
        super().__init__()

        # Setup private attributes
        self._wcs = None
        self._mapping = None

        # Lookup tables is a list of (pixel_dim, LookupTableCoord) to allow for
        # one pixel dimension having more than one lookup coord.
        self._lookup_tables = []
        self._dropped_tables = []

        # We need a reference to the parent NDCube
        self._ndcube = ndcube

    @classmethod
    def from_lookup_tables(cls, names, pixel_dimensions, lookup_tables, physical_types=None):
        """
        Construct a new ExtraCoords instance from lookup tables.

        This is a convenience wrapper around `ndcube.ExtraCoords.add` which does not
        expose all the options available in that method.

        Parameters
        ----------
        names : `tuple` of `str`
            The names of the world coordinates.
        pixel_dimensions : `tuple` of `int`
            The pixel dimensions (in the array) to which the ``lookup_tables``
            apply. Must be the same length as ``lookup_tables``.
        lookup_tables : iterable of `object`
            The lookup tables which specify the world coordinates for the ``pixel_dimensions``.
            Must be `~ndcube.extra_coords.BaseTableCoordinate` subclass instances or objects from
            which to instantiate them (see `ndcube.ExtraCoords.add`).
        physical_types: sequence of `str` or of sequences of `str`, optional
            Descriptors of the :ref:`dimensions`
            associated with each axis in the tables. Must be the same length as ``lookup_tables``;
            and length of each element must match the number of dimensions in corresponding
            ``lookup_tables[i]``.
            Defaults to `None`.

        Returns
        -------
        `ndcube.ExtraCoords`

        """
        if len(pixel_dimensions) != len(lookup_tables):
            raise ValueError(
                "The length of pixel_dimensions and lookup_tables must match."
            )

        if physical_types is None:
            physical_types = len(lookup_tables) * [physical_types]
        elif len(physical_types) != len(lookup_tables):
            raise ValueError("The number of physical types and lookup_tables must match.")

        extra_coords = cls()

        for name, pixel_dim, lookup_table, physical_type in zip(names, pixel_dimensions,
                                                                lookup_tables, physical_types):
            extra_coords.add(name, pixel_dim, lookup_table, physical_types=physical_type)

        return extra_coords

    def add(self, name, array_dimension, lookup_table, physical_types=None, **kwargs):
        # docstring in ABC

        if self._wcs is not None:
            raise ValueError(
                "Can not add a lookup_table to an ExtraCoords which was instantiated with a WCS object."
            )

        kwargs['names'] = [name] if not isinstance(name, (list, tuple)) else name

        if isinstance(lookup_table, BaseTableCoordinate):
            coord = lookup_table
        elif isinstance(lookup_table, Time):
            coord = TimeTableCoordinate(lookup_table, physical_types=physical_types, **kwargs)
        elif isinstance(lookup_table, SkyCoord):
            coord = SkyCoordTableCoordinate(lookup_table, physical_types=physical_types, **kwargs)
        elif isinstance(lookup_table, (list, tuple)):
            coord = QuantityTableCoordinate(*lookup_table, physical_types=physical_types, **kwargs)
        elif isinstance(lookup_table, u.Quantity):
            coord = QuantityTableCoordinate(lookup_table, physical_types=physical_types, **kwargs)
        else:
            raise TypeError(f"The input type {type(lookup_table)} isn't supported")

        self._lookup_tables.append((array_dimension, coord))

        # Sort the LUTs so that the mapping and the wcs are ordered in pixel dim order
        self._lookup_tables = sorted(self._lookup_tables,
                                          key=lambda x: x[0] if isinstance(x[0], Integral) else x[0][0])

    @property
    def _name_lut_map(self):
        """
        Map of world names to the corresponding `.LookupTableCoord`
        """
        return {lut[1].wcs.world_axis_names: lut for lut in self._lookup_tables}

    def keys(self):
        # docstring in ABC
        if not self.wcs:
            return ()

        return tuple(self.wcs.world_axis_names) if self.wcs.world_axis_names else None

    @property
    def mapping(self):
        # docstring in ABC
        if self._mapping:
            return self._mapping

        # If mapping is not set but lookup_tables is empty then the extra
        # coords is empty, so there is no mapping.
        if not self._lookup_tables:
            return ()

        # The mapping is from the array index (position in the list) to the
        # pixel dimensions (numbers in the list)
        lts = [list([lt[0]] if isinstance(lt[0], Integral) else lt[0]) for lt in self._lookup_tables]
        converter = partial(convert_between_array_and_pixel_axes, naxes=len(self._ndcube.shape))
        pixel_indicies = [list(converter(np.array(ids))) for ids in lts]
        return tuple(reduce(list.__add__, pixel_indicies))

    @mapping.setter
    def mapping(self, mapping):
        if self._mapping is not None:
            raise AttributeError("Can't set mapping if a mapping has already been specified.")

        if self._lookup_tables:
            raise AttributeError(
                "Can't set mapping manually when ExtraCoords is built from lookup tables."
            )

        if self._wcs is not None:
            if not max(mapping) <= self._wcs.pixel_n_dim - 1:
                raise ValueError(
                    "Values in the mapping can not be larger than the number of pixel dimensions in the WCS."
                )

        self._mapping = mapping

    @property
    def wcs(self):
        # docstring in ABC
        if self._wcs is not None:
            return self._wcs

        if not self._lookup_tables:
            return None

        tcoords = {lt[1] for lt in self._lookup_tables}
        # created a sorted list of unique items
        _tmp = set()  # a temporary set
        tcoords = [x[1] for x in self._lookup_tables if x[1] not in _tmp and _tmp.add(x[1]) is None]
        return MultipleTableCoordinate(*tcoords).wcs

    @wcs.setter
    def wcs(self, wcs):
        if self._wcs is not None:
            raise AttributeError(
                "Can't set wcs if a WCS has already been specified."
            )

        if self._lookup_tables:
            raise AttributeError(
                "Can't set wcs manually when ExtraCoords is built from lookup tables."
            )

        if self._mapping is not None:
            if not max(self._mapping) <= wcs.pixel_n_dim - 1:
                raise ValueError(
                    "Values in the mapping can not be larger than the number of pixel dimensions in the WCS."
                )

        self._wcs = wcs

    @property
    def is_empty(self):
        # docstring in ABC
        if not self._wcs and not self._lookup_tables:
            return True
        return False

    def _getitem_string(self, item):
        """
        Slice the Extracoords based on axis names.
        """

        for names, lut in self._name_lut_map.items():
            if item in names:
                new_ec = ExtraCoords(ndcube=self._ndcube)
                new_ec._lookup_tables = [lut]
                return new_ec

        raise KeyError(f"Can't find the world axis named {item} in this ExtraCoords object.")

    def _getitem_lookup_tables(self, item):
        """
        Apply an array slice to the lookup tables.

        Returns a new ExtraCoords object with modified lookup tables.
        """
        dropped_tables = set()
        new_lookup_tables = set()
        ndims = max([lut[0] if isinstance(lut[0], Integral) else max(lut[0])
                     for lut in self._lookup_tables]) + 1
        # Determine how many dimensions will be dropped by slicing below each dimension.
        if isinstance(item, Integral):
            n_dropped_dims = np.ones(ndims, dtype=int)
            item = tuple([item] + [slice(None)] * (ndims - 1))
        elif isinstance(item, slice):
            n_dropped_dims = np.zeros(ndims, dtype=int)
            item = tuple([item] + [slice(None)] * (ndims - 1))
        else:
            item = list(item) + [slice(None)] * (ndims - len(item))
            n_dropped_dims = np.cumsum([isinstance(i, Integral) for i in item])
        for lut_axis, lut in self._lookup_tables:
            lut_axes = (lut_axis,) if not isinstance(lut_axis, tuple) else lut_axis
            new_lut_axes = tuple(ax - n_dropped_dims[ax] for ax in lut_axes)
            lut_slice = tuple(item[i] for i in lut_axes)
            if isinstance(lut_slice, tuple) and len(lut_slice) == 1:
                lut_slice = lut_slice[0]

            sliced_lut = lut[lut_slice]

            if sliced_lut.is_scalar():
                dropped_tables.add(sliced_lut)
            else:
                new_lookup_tables.add((new_lut_axes, sliced_lut))
        new_extra_coords = type(self)()
        new_extra_coords._lookup_tables = list(new_lookup_tables)
        new_extra_coords._dropped_tables = list(dropped_tables)
        return new_extra_coords

    def _getitem_wcs(self, item):
        item = sanitize_slices(item, self.wcs.pixel_n_dim)

        # It's valid to slice down the EC such that there is nothing left,
        # which is not a valid way to slice the WCS
        if len(item) == self.wcs.pixel_n_dim and all(isinstance(i, Integral) for i in item):
            return type(self)()

        subwcs = self.wcs[item]

        new_mapping = [self.mapping[i] for i, subitem in enumerate(item) if not isinstance(subitem, Integral)]

        new_ec = type(self)()
        new_ec.wcs = subwcs
        new_ec.mapping = new_mapping
        return new_ec

    def __getitem__(self, item):
        # docstring in ABC
        if isinstance(item, str):
            return self._getitem_string(item)

        if self._wcs:
            return self._getitem_wcs(item)

        if self._lookup_tables:
            return self._getitem_lookup_tables(item)

        # If we get here this object is empty, so just return an empty extra coords
        # This is done to simplify the slicing in NDCube
        return self

    @property
    def dropped_world_dimensions(self):
        """
        Return an APE-14 like representation of any sliced out world dimensions.
        """

        if self._wcs:
            if isinstance(self._wcs, SlicedLowLevelWCS):
                return self._wcs.dropped_world_dimensions

        if self._lookup_tables or self._dropped_tables:
            mtc = MultipleTableCoordinate(*[lt[1] for lt in self._lookup_tables])
            mtc._dropped_coords = self._dropped_tables

            return mtc.dropped_world_dimensions

        return {}

    def resample(self, factor, offset=0, ndcube=None, **kwargs):
        """
        Resample all extra coords by given factors in array-index-space.

        One resample factor must be supplied for each array axis in array-axis order.

        Parameters
        ----------
        factor: `int`, `float`, or iterable thereof.
            The factor by which each array axis is resampled.
            If scalar, same factor is applied to all axes.
            Otherwise a factor for each axis must be provided.

        offset: `int`, `float`, or iterable therefore.
            The location on the underlying grid which corresponds
            to the zeroth element after resampling. If iterable, must have an entry
            for each dimension. If a scalar, the grid will be
            shifted by the same amount in all dimensions.

        ndcube: `~ndcube.NDCube`
            The NDCube instance with which the output ExtraCoords object is associated.

        kwargs
            All remaining kwargs are passed to `numpy.interp`.

        Returns
        -------
        new_ec: `~ndcube.ExtraCoords`
            A new ExtraCoords object holding the interpolated coords.
        """
        new_ec = type(self)(ndcube)
        if self.is_empty:
            return new_ec
        if self._ndcube is not None:
            cube_shape = self._ndcube.data.shape
            ndim = len(cube_shape)
        elif self._wcs is not None:
            ndim = self._wcs.pixel_n_dim
        else:
            raise NotImplementedError(
                "Resampling a lookup-table-based ExtraCoords not yet implemented. "
                "Please raise an issue at https://github.com/sunpy/ndcube/issues "
                "if you need this functionality")
        if np.isscalar(factor):
            factor = [factor] * ndim
        if len(factor) != ndim:
            raise ValueError(
                "factor must be scalar or an iterable with length equal to number of cube "
                f"dimensions: len(factor) = {len(factor)}; No. cube dimensions = {ndim}.")
        if np.isscalar(offset):
            offset = [offset] * ndim
        if len(offset) != ndim:
            raise ValueError(
                "offset must be scalar or an iterable with length equal to number of cube "
                f"dimensions: len(offset) = {len(offset)}; No. cube dimensions = {ndim}.")
        # If ExtraCoords object built on WCS, resample using WCS insfrastructure
        if self._wcs is not None:
            new_ec.wcs = HighLevelWCSWrapper(ResampledLowLevelWCS(self._wcs.low_level_wcs,
                                                                  factor, offset))
            return new_ec
        # Else interpolate the lookup table coordinates.
        factor = np.asarray(factor)
        new_grids = []
        for c, d, f in zip(offset, cube_shape, factor):
            x = np.arange(c, d+f, f)
            x = x[x <= d-1]
            new_grids.append(x)
        new_grids = np.array(new_grids, dtype=object)
        for array_axes, coord in self._lookup_tables:
            if np.isscalar(array_axes):
                new_coord = coord.interpolate(new_grids[array_axes], **kwargs)
            else:
                new_coord = coord.interpolate(*new_grids[np.asarray(array_axes)], **kwargs)
            new_ec.add(coord.names, array_axes, new_coord, physical_types=coord.physical_types)
        return new_ec

    @property
    def cube_wcs(self):
        """Produce a WCS that describes the associated NDCube with just the extra coords.

        For NDCube pixel axes without any extra coord, dummy axes are inserted.
        """
        wcses = [self.wcs]
        mapping = list(self.mapping)
        dummy_axes = self._cube_array_axes_without_extra_coords
        n_dummy_axes = len(dummy_axes)
        if n_dummy_axes > 0:
            dummy_wcs = WCS(naxis=n_dummy_axes)
            dummy_wcs.wcs.crpix = [1] * n_dummy_axes
            dummy_wcs.wcs.cdelt = [1] * n_dummy_axes
            dummy_wcs.wcs.crval = [0] * n_dummy_axes
            dummy_wcs.wcs.ctype = ["PIXEL"] * n_dummy_axes
            dummy_wcs.wcs.cunit = ["pix"] * n_dummy_axes
            wcses.append(dummy_wcs)
            mapping += list(dummy_axes)
        return CompoundLowLevelWCS(*wcses, mapping=mapping)

    @property
    def _cube_array_axes_without_extra_coords(self):
        """Return the array axes not associated with any extra coord."""
        return set(range(len(self._ndcube.shape))) - set(self.mapping)

    def __str__(self):
        classname = self.__class__.__name__
        elements = [f"{', '.join(table.names)} ({axes}) {table.physical_types}: {table}"
                    for axes, table in self._lookup_tables]
        length = len(classname) + 2 * len(elements) + sum(len(e) for e in elements)
        if length > np.get_printoptions()['linewidth']:
            joiner = ',\n ' + len(classname) * ' '
        else:
            joiner = ', '

        return f"{classname}({joiner.join(elements)})"

    def __repr__(self):
        return f"{object.__repr__(self)}\n{self}"



"""
===============================================
How to create an GWCS from quantities and times
===============================================

This example shows how to create a GWCS from astropy quantities.
"""
import numpy as np
from matplotlib import pyplot as plt

import astropy.units as u
from astropy.time import Time

from ndcube import NDCube
from ndcube.extra_coords import QuantityTableCoordinate, TimeTableCoordinate

##############################################################################
# We aim to create coordinates that are focused around time and energies using astropy quantities.

energy = np.arange(10) * u.keV
time = Time('2020-01-01 00:00:00') + np.arange(9)*u.s

##############################################################################
# Then, we need to turn these into lookup tables using
# `~ndcube.extra_coords.table_coord.QuantityTableCoordinate` and
# `~ndcube.extra_coords.table_coord.TimeTableCoordinate` to create table coordinates.

energy_coord = QuantityTableCoordinate(energy, names='energy', physical_types='em.energy')
print(energy_coord)

time_coord = TimeTableCoordinate(time, names='time', physical_types='time')
print(time_coord)

##############################################################################
# Now we need to combine table coordinates created above and extract the ``.wcs`` from the result.

wcs = (time_coord & energy_coord).wcs
print(wcs)

##############################################################################
# Now, we have all the pieces required to construct a `~ndcube.NDCube` with this data and the GWCS we just created.

data = np.random.rand(len(time), len(energy))
cube = NDCube(data=data, wcs=wcs)
print(cube)

##############################################################################
# Finally, we will plot the cube.

cube.plot()

plt.show()



#
# Configuration file for the Sphinx documentation builder.
import os
import warnings
import datetime
from pathlib import Path

from astropy.utils.exceptions import AstropyDeprecationWarning
from matplotlib import MatplotlibDeprecationWarning
from packaging.version import Version

# -- Read the Docs Specific Configuration ------------------------------------
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    os.environ['HIDE_PARFIVE_PROGESS'] = 'True'

# -- Project information -----------------------------------------------------

# The full version, including alpha/beta/rc tags
from ndcube import __version__

_version = Version(__version__)
version = release = str(_version)
# Avoid "post" appearing in version string in rendered docs
if _version.is_postrelease:
    version = release = _version.base_version
# Avoid long githashes in rendered Sphinx docs
elif _version.is_devrelease:
    version = release = f"{_version.base_version}.dev{_version.dev}"
is_development = _version.is_devrelease
is_release = not (_version.is_prerelease or _version.is_devrelease)

project = "ndcube"
author = "The SunPy Community"
copyright = f'{datetime.datetime.now().year}, {author}'  # noqa: A001

warnings.filterwarnings("error", category=MatplotlibDeprecationWarning)
warnings.filterwarnings("error", category=AstropyDeprecationWarning)

# -- General configuration ---------------------------------------------------

extensions = [
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.doctest',
    'ndcube.utils.sphinx.code_context',
    'sphinx_gallery.gen_gallery',
    "sphinxext.opengraph",
    "sphinx.ext.mathjax",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "sphinx_changelog",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# Treat everything in single ` as a Python reference.
default_role = "py:obj"

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/',
               (None, 'http://data.astropy.org/intersphinx/python3.inv')),
    'numpy': ('https://docs.scipy.org/doc/numpy/',
              (None, 'http://data.astropy.org/intersphinx/numpy.inv')),
    'matplotlib': ('https://matplotlib.org/',
                   (None, 'http://data.astropy.org/intersphinx/matplotlib.inv')),
    'astropy': ('http://docs.astropy.org/en/stable/', None),
    'sunpy': ('https://docs.sunpy.org/en/stable/', None),
    'mpl_animators': ('https://docs.sunpy.org/projects/mpl-animators/en/stable/', None),
    'gwcs': ('https://gwcs.readthedocs.io/en/stable/', None),
    'reproject': ("https://reproject.readthedocs.io/en/stable/", None)
    }

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sunpy"

html_logo = png_icon = 'logo/ndcube.png'

html_favicon = 'logo/favicon.png'

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# By default, when rendering docstrings for classes, sphinx.ext.autodoc will
# make docs with the class-level docstring and the class-method docstrings,
# but not the __init__ docstring, which often contains the parameters to
# class constructors across the scientific Python ecosystem. The option below
# will append the __init__ docstring to the class-level docstring when rendering
# the docs. For more options, see:
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autoclass_content
autoclass_content = "both"

# -- Other options ----------------------------------------------------------

napoleon_use_rtype = False
napoleon_google_docstring = False
napoleon_use_param = False

nitpicky = True
# This is not used. See docs/nitpick-exceptions file for the actual listing.
nitpick_ignore = []
for line in open('nitpick-exceptions'):
    if line.strip() == "" or line.startswith("#"):
        continue
    dtype, target = line.split(None, 1)
    target = target.strip()
    nitpick_ignore.append((dtype, target))

# -- Sphinx Gallery ---------------------------------------------------------

sphinx_gallery_conf = {
    'backreferences_dir': Path('generated/modules'),
    'filename_pattern': '^((?!skip_).)*$',
    'examples_dirs': Path('../examples'),
    'within_subsection_order': "ExampleTitleSortKey",
    'gallery_dirs': Path('generated/gallery'),
    'matplotlib_animations': True,
    "default_thumb_file": png_icon,
    'abort_on_example_error': False,
    'plot_gallery': 'True',
    'remove_config_comments': True,
    'doc_module': ('ndcube'),
    'only_warn_on_example_error': True,
}

# -- Sphinxext Opengraph ----------------------------------------------------

ogp_image = "https://github.com/sunpy/ndcube/raw/main/docs/logo/ndcube.png"
ogp_use_first_image = True
ogp_description_length = 160
ogp_custom_meta_tags = [
    '<meta property="og:ignore_canonical" content="true" />',
]



from numbers import Integral

import numpy as np

from astropy.wcs import WCS
from astropy.wcs.wcsapi import SlicedLowLevelWCS
from astropy.wcs.wcsapi.wrappers.base import BaseWCSWrapper

from ndcube.wcs.wrappers import ResampledLowLevelWCS

__all__ = ["unwrap_wcs_to_fitswcs"]


def unwrap_wcs_to_fitswcs(wcs):
    """
    Create FITS-WCS equivalent to (nested) WCS wrapper object.

    Underlying WCS must be FITS-WCS.
    No axes are dropped from original FITS-WCS, even if sliced by an integer.
    Instead, integer-sliced axes is sliced to length-1 and marked True in the
    ``dropped_data_axes`` output.
    Currently supported wrapper classes include `astropy.wcs.wcsapi.SlicedLowLevelWCS`
    and `ndcube.wcs.wrappers.ResampledLowLevelWCS`.

    Parameters
    ----------
    wcs: `~astropy.wcs.wcsapi.BaseWCSWrapper`
        The WCS Wrapper object.
        Base level WCS implementation must be FITS-WCS.

    Returns
    -------
    fitswcs: `astropy.wcs.WCS`
        The equivalent FITS-WCS object.
    dropped_data_axes: 1-D `numpy.ndarray`
        Denotes which axes must have been dropped from the data array by slicing wrappers.
        Axes are in array/numpy order, reversed compared to WCS.
    """
    # If wcs is already a FITS-WCS, return it.
    low_level_wrapper = wcs.low_level_wcs if hasattr(wcs, "low_level_wcs") else wcs
    if isinstance(low_level_wrapper, WCS):
        return low_level_wrapper, np.zeros(low_level_wrapper.naxis, dtype=bool)
    # Determine chain of wrappers down to the FITS-WCS.
    wrapper_chain = []
    while isinstance(low_level_wrapper, BaseWCSWrapper):
        wrapper_chain.append(low_level_wrapper)
        low_level_wrapper = low_level_wrapper._wcs
        if hasattr(low_level_wrapper, "low_level_wcs"):
            low_level_wrapper = low_level_wrapper.low_level_wcs
    if not isinstance(low_level_wrapper, WCS):
        raise TypeError(f"Base-level WCS must be type {type(WCS)}. Found: {type(low_level_wrapper)}")
    fitswcs = low_level_wrapper
    dropped_data_axes = np.zeros(fitswcs.naxis, dtype=bool)
    # Unwrap each wrapper in reverse order and edit fitswcs.
    for low_level_wrapper in wrapper_chain[::-1]:
        if isinstance(low_level_wrapper, SlicedLowLevelWCS):
            slice_items = np.array([slice(None)] * fitswcs.naxis)
            slice_items[dropped_data_axes == False] = low_level_wrapper._slices_array  # numpy order  # NOQA: E712
            fitswcs, dda = _slice_fitswcs(fitswcs, slice_items, numpy_order=True)
            dropped_data_axes[dda] = True
        elif isinstance(low_level_wrapper, ResampledLowLevelWCS):
            factor = np.ones(fitswcs.naxis)
            offset = np.zeros(fitswcs.naxis)
            kept_wcs_axes = dropped_data_axes[::-1] == False  # WCS-order  # NOQA: E712
            factor[kept_wcs_axes] = low_level_wrapper._factor
            offset[kept_wcs_axes] = low_level_wrapper._offset
            fitswcs = _resample_fitswcs(fitswcs, factor, offset)
        else:
            raise TypeError("Unrecognized/unsupported WCS Wrapper type: {type(low_level_wrapper)}")
    return fitswcs, dropped_data_axes


def _slice_fitswcs(fitswcs, slice_items, numpy_order=True, shape=None):
    """
    Slice a FITS-WCS.

    If an `int` is given in ``slice_items``, the corresponding axis is not dropped.
    But the new 0th pixel will correspond the index given by the `int` in the
    original WCS.

    Parameters
    ----------
    fitswcs: `astropy.wcs.WCS`
        The FITS-WCS object to be sliced.
    slice_items: iterable of `slice` objects or `int`
        The slices to by applied to each axis. If an `int` is provided, the axis
        is sliced to length-1, but not dropped. However, its corresponding entry
        in the ``dropped_data_axes`` output is marked True.
    numpy_order: `bool`
        If True, slices in ``slice_items`` are in array/numpy order, which is
        reversed compared to the WCS order.
    shape: sequence of `int`, optional
        The length of each axis.  Only used if negative indices are supplied
        in ``slice_items``.  If not supplied, set to ``fitswcs._naxis``.
        Order defined by numpy_order kwarg.

    Returns
    -------
    sliced_wcs: `astropy.wcs.WCS`
        The sliced FITS-WCS.
    dropped_data_axes: 1-D `numpy.ndarray`
        Denotes which axes must have been dropped from the data array by slicing wrappers.
        Order of axes (numpy or WCS) is dictated by ``numpy_order`` kwarg.
    """
    def negative_index_error_msg(x): return (
        f"Negative indexing not supported as {x}th axis length is 0 in "
        "underlying FITS-WCS. Supply axes lengths via shape kwarg.")
    naxis = fitswcs.naxis
    dropped_data_axes = np.zeros(naxis, dtype=bool)
    # Sanitize inputs
    if shape is None:
        shape = fitswcs._naxis
        if numpy_order:
            shape = shape[::-1]
    else:
        if len(shape) != naxis:
            raise ValueError("shape kwarg must be same length as number of pixel axes "
                             f"in FITS-WCS, i.e. {naxis}")
        if not all(isinstance(s, Integral) for s in shape):
            raise TypeError("All elements of ``shape`` must be integers. "
                            f"shapes types = {[type(s) for s in shape]}")
    slice_items = list(slice_items)
    for i, (item, len_axis) in enumerate(zip(slice_items, shape)):
        if isinstance(item, Integral):
            # Mark axis corresponding to int item as dropped from data array.
            dropped_data_axes[i] = True
            # Convert negative indices to positive equivalent.
            if item < 0:
                if len_axis == 0:
                    raise ValueError(negative_index_error_msg(i))
                item = len_axis + item
            # Convert int item to slice so a FITS-WCS is returned after slicing.
            slice_items[i] = slice(item, item + 1)
        elif isinstance(item, slice):
            # Convert negative indices inside slice item to positive equivalent.
            start_neg = item.start is not None and item.start < 0
            stop_neg = item.stop is not None and item.stop < 0
            if start_neg or stop_neg:
                if len_axis == 0:
                    raise ValueError(negative_index_error_msg(i))
                start = len_axis + item.start if start_neg else item.start
                stop = len_axis + item.stop if stop_neg else item.stop
                slice_items[i] = slice(start, stop, item.step)
        else:
            raise TypeError("All slice_items must be a slice or an int. "
                            f"type(slice_items[{i}]) = {type(slice_items[i])}")
    # Slice WCS
    sliced_wcs = fitswcs.slice(slice_items, numpy_order=numpy_order)
    return sliced_wcs, dropped_data_axes


def _resample_fitswcs(fitswcs, factor, offset=0):
    """
    Resample the plate scale of a FITS-WCS by a given factor.

    ``factor`` and ``offset`` inputs are in pixel order.

    Parameters
    ----------
    fitswcs: `astropy.wcs.WCS`
        The FITS-WCS object to be resampled.
    factor: 1-D array-like or scalar
        The factor by which the FITS-WCS is resampled.
        Must be same length as number of axes in ``fitswcs``.
        If scalar, the same factor is applied to all axes.
        Factors must be given in WCS-order (opposite to data axes order).
    offset: 1-D array-like or scalar
        The location on the initial pixel grid which corresponds to zero on the
        resampled pixel grid. If scalar, the same offset is applied to all axes.
        Offsets must be given in WCS-order (opposite to data axes order).

    Returns
    -------
    resampled_wcs: `astropy.wcs.WCS`
        The resampled FITS-WCS.
    """
    # Sanitize inputs.
    factor = np.asarray(factor)
    if len(factor) != fitswcs.naxis:
        raise ValueError(f"Length of factor must equal number of dimensions {fitswcs.naxis}.")
    offset = np.asarray(offset)
    if len(offset) != fitswcs.naxis:
        raise ValueError(f"Length of offset must equal number of dimensions {fitswcs.naxis}.")
    # Scale plate scale and shift by offset.
    fitswcs.wcs.cdelt *= factor
    fitswcs.wcs.crpix = (fitswcs.wcs.crpix + offset) / factor
    fitswcs._naxis = list(np.round(np.array(fitswcs._naxis) / factor).astype(int))
    return fitswcs



"""
=======================================================
How to create an NDCube from data stored in a FITS file
=======================================================

This example shows how you load in data from a FITS file to create an `~ndcube.NDCube`.
Here we will use an example of a single image.
"""

import matplotlib.pyplot as plt

from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS

from ndcube import NDCube

##############################################################################
# We first download the example file that we will use here to show how to create an
# `~ndcube.NDCube` from data stored in a FITS file.
# Here we are using an example file from ``astropy``.

image_file = get_pkg_data_filename('tutorials/FITS-images/HorseHead.fits')

###########################################################################
# Lets extract the image data and the header information from the FITS file.
# This can be achieved by using the functionality within `astropy.io.fits`.
# In this file the image information is located in the Primary HDU (extension 0).

image_data = fits.getdata(image_file)
image_header = fits.getheader(image_file)

##########################################################################
# To create an `~ndcube.NDCube` object, we need both the data array and a
# WCS object (e.g. an `~astropy.wcs.WCS`). Here the data WCS information is
# within the header, which we can pass to :class:`~astropy.wcs.WCS` to create a WCS object.

example_ndcube = NDCube(image_data, WCS(image_header))

##########################################################################
# Now we have created an `~ndcube.NDCube` from this data.
# We can inspect the `~ndcube.NDCube`, such as the WCS.

print(example_ndcube.wcs)

##########################################################################
# and we can also inspect the dimensions.

print(example_ndcube.shape)

##########################################################################
# We can also quickly visualize the data using the :meth:`~ndcube.NDCube.plot` method.

example_ndcube.plot()

plt.show()



"""
This file contains a set of common fixtures to get a set of different but
predictable NDCube objects.
"""
import logging

import dask.array
import numpy as np
import pytest

import astropy.nddata
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.nddata import StdDevUncertainty
from astropy.time import Time, TimeDelta
from astropy.wcs import WCS

from ndcube import ExtraCoords, GlobalCoords, NDCube, NDCubeSequence, NDMeta
from ndcube.tests import helpers

# Force MPL to use non-gui backends for testing.
try:
    import matplotlib
    import matplotlib.pyplot as plt
except ImportError:
    HAVE_MATPLOTLIB = False
else:
    HAVE_MATPLOTLIB = True
    matplotlib.use('Agg')


console_logger = logging.getLogger()
console_logger.setLevel('INFO')

################################################################################
# Helper Functions
################################################################################


def skycoord_2d_lut(shape):
    total_len = np.prod(shape)
    data = (np.arange(total_len).reshape(shape),
            np.arange(total_len, total_len * 2).reshape(shape))
    return SkyCoord(*data, unit=u.deg)


def data_nd(shape, dtype=float):
    nelem = np.prod(shape)
    return np.arange(nelem, dtype=dtype).reshape(shape)


def time_extra_coords(shape, axis, base):
    return ExtraCoords.from_lookup_tables(
        ('time',),
        (axis,),
        (base + TimeDelta([i * 60 for i in range(shape[axis])], format='sec'),))


def gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l, time_axis, time_base, global_coords=None):
    shape = (10, 5, 8)
    wcs_3d_lt_ln_l.array_shape = shape
    data_cube = data_nd(shape)
    mask = data_cube < 0
    meta = {"message": "hello world"}
    unit = u.ph
    extra_coords = time_extra_coords(shape, time_axis, time_base)
    cube = NDCube(data_cube,
                  wcs_3d_lt_ln_l,
                  mask=mask,
                  uncertainty=data_cube,
                  meta=meta,
                  unit=unit)
    cube._extra_coords = extra_coords

    if global_coords:
        cube._global_coords = global_coords

    return cube


################################################################################
# WCS Fixtures
################################################################################


@pytest.fixture
def wcs_4d_t_l_lt_ln():
    header = {
        'CTYPE1': 'TIME    ',
        'CUNIT1': 'min',
        'CDELT1': 0.4,
        'CRPIX1': 0,
        'CRVAL1': 0,

        'CTYPE2': 'WAVE    ',
        'CUNIT2': 'Angstrom',
        'CDELT2': 0.2,
        'CRPIX2': 0,
        'CRVAL2': 0,

        'CTYPE3': 'HPLT-TAN',
        'CUNIT3': 'arcsec',
        'CDELT3': 20,
        'CRPIX3': 0,
        'CRVAL3': 0,

        'CTYPE4': 'HPLN-TAN',
        'CUNIT4': 'arcsec',
        'CDELT4': 5,
        'CRPIX4': 5,
        'CRVAL4': 0,

        'DATEREF': "2020-01-01T00:00:00"
    }
    return WCS(header=header)


@pytest.fixture
def wcs_4d_lt_t_l_ln():
    header = {
        'CTYPE1': 'HPLT-TAN',
        'CUNIT1': 'arcsec',
        'CDELT1': 20,
        'CRPIX1': 0,
        'CRVAL1': 0,

        'CTYPE2': 'TIME    ',
        'CUNIT2': 'min',
        'CDELT2': 0.4,
        'CRPIX2': 0,
        'CRVAL2': 0,

        'CTYPE3': 'WAVE    ',
        'CUNIT3': 'Angstrom',
        'CDELT3': 0.2,
        'CRPIX3': 0,
        'CRVAL3': 0,

        'CTYPE4': 'HPLN-TAN',
        'CUNIT4': 'arcsec',
        'CDELT4': 5,
        'CRPIX4': 5,
        'CRVAL4': 0,

        'DATEREF': "2020-01-01T00:00:00"
    }
    return WCS(header=header)


@pytest.fixture
def wcs_3d_l_lt_ln():
    header = {
        'CTYPE1': 'WAVE    ',
        'CUNIT1': 'Angstrom',
        'CDELT1': 0.2,
        'CRPIX1': 0,
        'CRVAL1': 10,

        'CTYPE2': 'HPLT-TAN',
        'CUNIT2': 'arcsec',
        'CDELT2': 5,
        'CRPIX2': 5,
        'CRVAL2': 0,

        'CTYPE3': 'HPLN-TAN',
        'CUNIT3': 'arcsec',
        'CDELT3': 10,
        'CRPIX3': 0,
        'CRVAL3': 0,
    }

    return WCS(header=header)


@pytest.fixture
def wcs_3d_lt_ln_l():
    header = {

        'CTYPE1': 'HPLN-TAN',
        'CUNIT1': 'arcsec',
        'CDELT1': 10,
        'CRPIX1': 0,
        'CRVAL1': 0,

        'CTYPE2': 'HPLT-TAN',
        'CUNIT2': 'arcsec',
        'CDELT2': 5,
        'CRPIX2': 5,
        'CRVAL2': 0,

        'CTYPE3': 'WAVE    ',
        'CUNIT3': 'Angstrom',
        'CDELT3': 0.2,
        'CRPIX3': 0,
        'CRVAL3': 10,
    }

    return WCS(header=header)


@pytest.fixture
def wcs_3d_wave_lt_ln():
    header = {
        'CTYPE1': 'WAVE    ',
        'CUNIT1': 'Angstrom',
        'CDELT1': 0.2,
        'CRPIX1': 0,
        'CRVAL1': 10,

        'CTYPE2': 'HPLT-TAN',
        'CUNIT2': 'deg',
        'CDELT2': 0.5,
        'CRPIX2': 2,
        'CRVAL2': 0.5,

        'CTYPE3': 'HPLN-TAN    ',
        'CUNIT3': 'deg',
        'CDELT3': 0.4,
        'CRPIX3': 2,
        'CRVAL3': 1,
    }
    return WCS(header=header)


@pytest.fixture
def wcs_2d_lt_ln():
    spatial = {
        'CTYPE1': 'HPLT-TAN',
        'CUNIT1': 'arcsec',
        'CDELT1': 2,
        'CRPIX1': 5,
        'CRVAL1': 0,

        'CTYPE2': 'HPLN-TAN',
        'CUNIT2': 'arcsec',
        'CDELT2': 4,
        'CRPIX2': 5,
        'CRVAL2': 0,
    }
    return WCS(header=spatial)


@pytest.fixture
def wcs_1d_l():
    spatial = {
        'CNAME1': 'spectral',
        'CTYPE1': 'WAVE',
        'CUNIT1': 'nm',
        'CDELT1': 0.5,
        'CRPIX1': 2,
        'CRVAL1': 0.5,
    }
    return WCS(header=spatial)


@pytest.fixture
def wcs_3d_ln_lt_t_rotated():
    h_rotated = {
        'CTYPE1': 'HPLN-TAN',
        'CUNIT1': 'arcsec',
        'CDELT1': 0.4,
        'CRPIX1': 0,
        'CRVAL1': 0,
        'NAXIS1': 5,

        'CTYPE2': 'HPLT-TAN',
        'CUNIT2': 'arcsec',
        'CDELT2': 0.5,
        'CRPIX2': 0,
        'CRVAL2': 0,
        'NAXIS2': 5,

        'CTYPE3': 'TIME    ',
        'CUNIT3': 's',
        'CDELT3': 3,
        'CRPIX3': 0,
        'CRVAL3': 0,
        'NAXIS3': 2,

        'DATEREF': "2020-01-01T00:00:00",

        'PC1_1': 0.714963912964,
        'PC1_2': -0.699137151241,
        'PC1_3': 0.0,
        'PC2_1': 0.699137151241,
        'PC2_2': 0.714963912964,
        'PC2_3': 0.0,
        'PC3_1': 0.0,
        'PC3_2': 0.0,
        'PC3_3': 1.0
    }
    return WCS(header=h_rotated)


@pytest.fixture
def wcs_3d_ln_lt_l_coupled():
    # WCS for a 3D data cube with two celestial axes and one wavelength axis.
    # The latitudinal dimension is coupled to the third pixel dimension through
    # a single off diagonal element in the PCij matrix
    header = {
        'CTYPE1': 'HPLN-TAN',
        'CRPIX1': 5,
        'CDELT1': 5,
        'CUNIT1': 'arcsec',
        'CRVAL1': 0.0,

        'CTYPE2': 'HPLT-TAN',
        'CRPIX2': 5,
        'CDELT2': 5,
        'CUNIT2': 'arcsec',
        'CRVAL2': 0.0,

        'CTYPE3': 'WAVE',
        'CRPIX3': 1.0,
        'CDELT3': 1,
        'CUNIT3': 'Angstrom',
        'CRVAL3': 1.0,

        'PC1_1': 1,
        'PC1_2': 0,
        'PC1_3': 0,
        'PC2_1': 0,
        'PC2_2': 1,
        'PC2_3': -1.0,
        'PC3_1': 0.0,
        'PC3_2': 0.0,
        'PC3_3': 1.0,

        'WCSAXES': 3,

        'DATEREF': "2020-01-01T00:00:00"
    }
    return WCS(header=header)


@pytest.fixture
def wcs_3d_ln_lt_t_coupled():
    # WCS for a 3D data cube with two celestial axes and one time axis.
    header = {
        'CTYPE1': 'HPLN-TAN',
        'CRPIX1': 5,
        'CDELT1': 5,
        'CUNIT1': 'arcsec',
        'CRVAL1': 0.0,

        'CTYPE2': 'HPLT-TAN',
        'CRPIX2': 5,
        'CDELT2': 5,
        'CUNIT2': 'arcsec',
        'CRVAL2': 0.0,

        'CTYPE3': 'UTC',
        'CRPIX3': 1.0,
        'CDELT3': 1,
        'CUNIT3': 's',
        'CRVAL3': 1.0,

        'PC1_1': 1,
        'PC1_2': 0,
        'PC1_3': 0,
        'PC2_1': 0,
        'PC2_2': 1,
        'PC2_3': 0,
        'PC3_1': 0,
        'PC3_2': 1,
        'PC3_3': 1,

        'WCSAXES': 3,

        'DATEREF': "2020-01-01T00:00:00"
    }
    return WCS(header=header)


################################################################################
# Extra and Global Coords Fixtures
################################################################################


@pytest.fixture
def simple_extra_coords_3d():
    return ExtraCoords.from_lookup_tables(('time', 'hello', 'bye'),
                                          (0, 1, 2),
                                          (list(range(2)) * u.pix,
                                           list(range(3)) * u.pix,
                                           list(range(4)) * u.pix
                                           )
                                          )


@pytest.fixture
def time_and_simple_extra_coords_2d():
    return ExtraCoords.from_lookup_tables(("time", "hello"),
                                          (0, 1),
                                          (Time(["2000-01-01T12:00:00", "2000-01-02T12:00:00"],
                                                scale="utc", format="fits"),
                                           list(range(3)) * u.pix)
                                          )


@pytest.fixture
def extra_coords_3d():
    coord0 = Time(["2000-01-01T12:00:00", "2000-01-02T12:00:00"], scale="utc", format="fits")
    coord1 = list(range(3)) * u.pix
    coord2 = list(range(4)) * u.m
    return ExtraCoords.from_lookup_tables(('time', 'bye', 'hello'),
                                          (0, 1, 2),
                                          (coord0, coord1, coord2)
                                          )


@pytest.fixture
def extra_coords_sharing_axis():
    return ExtraCoords.from_lookup_tables(('hello', 'bye'),
                                          (1, 1),
                                          (list(range(3)) * u.m,
                                           list(range(3)) * u.keV,
                                           )
                                          )

################################################################################
# NDCube Fixtures
################################################################################

@pytest.fixture
def ndcube_4d_ln_l_t_lt(wcs_4d_lt_t_l_ln):
    shape = (5, 10, 12, 8)
    wcs_4d_lt_t_l_ln.array_shape = shape
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=wcs_4d_lt_t_l_ln)


@pytest.fixture
def ndcube_4d_ln_lt_l_t(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    wcs_4d_t_l_lt_ln.array_shape = shape
    data_cube = data_nd(shape, dtype=int)
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln)


@pytest.fixture
def ndcube_4d_axis_aware_meta(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    wcs_4d_t_l_lt_ln.array_shape = shape
    data_cube = data_nd(shape, dtype=int)
    meta = NDMeta({"a": "scalar",
                   "slit position": np.arange(shape[0], dtype=int),
                   "pixel label": np.arange(np.prod(shape[:2])).reshape(shape[:2]),
                   "line": ["Si IV"] * shape[2],
                   "exposure time": ([2] * shape[-1]) * u.s},
                  axes={"slit position": 0,
                        "pixel label": (0, 1),
                        "line": (2,),
                        "exposure time": 3})
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln, meta=meta)


@pytest.fixture
def ndcube_4d_uncertainty(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    data_cube = data_nd(shape)
    uncertainty = np.sqrt(data_cube)
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln, uncertainty=uncertainty)


@pytest.fixture
def ndcube_4d_mask(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    data_cube = data_nd(shape)
    uncertainty = np.sqrt(data_cube)
    mask = data_cube % 2
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln, uncertainty=uncertainty, mask=mask)


@pytest.fixture
def ndcube_4d_extra_coords(wcs_4d_t_l_lt_ln, simple_extra_coords_3d):
    shape = (5, 8, 10, 12)
    data_cube = data_nd(shape)
    cube = NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln)
    cube._extra_coords = simple_extra_coords_3d
    return cube


@pytest.fixture
def ndcube_4d_unit_uncertainty(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    data_cube = data_nd(shape)
    uncertainty = np.sqrt(data_cube)
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln,
                  unit=u.J, uncertainty=uncertainty)


@pytest.fixture
def ndcube_4d(request):
    """
    This is a meta fixture for parametrizing all the 4D ndcubes.
    """
    return request.getfixturevalue("ndcube_4d_" + request.param)


@pytest.fixture
def ndcube_3d_ln_lt_l(wcs_3d_l_lt_ln, simple_extra_coords_3d):
    shape = (2, 3, 4)
    wcs_3d_l_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_l_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    cube._extra_coords = simple_extra_coords_3d
    cube._extra_coords._ndcube = cube
    return cube


@pytest.fixture
def ndcube_3d_ln_lt_l_ec_all_axes(wcs_3d_l_lt_ln, extra_coords_3d):
    shape = (2, 3, 4)
    wcs_3d_l_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_l_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    cube._extra_coords = extra_coords_3d
    cube._extra_coords._ndcube = cube
    return cube


@pytest.fixture
def ndcube_3d_ln_lt_l_ec_sharing_axis(wcs_3d_l_lt_ln, extra_coords_sharing_axis):
    shape = (2, 3, 4)
    wcs_3d_l_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_l_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    cube._extra_coords = extra_coords_sharing_axis
    cube._extra_coords._ndcube = cube
    return cube


@pytest.fixture
def ndcube_3d_ln_lt_l_ec_time(wcs_3d_l_lt_ln, time_and_simple_extra_coords_2d):
    shape = (2, 3, 4)
    wcs_3d_l_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_l_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    cube._extra_coords = time_and_simple_extra_coords_2d
    cube._extra_coords._ndcube = cube
    return cube


@pytest.fixture
def ndcube_3d_wave_lt_ln_ec_time(wcs_3d_wave_lt_ln):
    shape = (3, 4, 5)
    wcs_3d_wave_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_wave_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    base_time = Time('2000-01-01', format='fits', scale='utc')
    timestamps = Time([base_time + TimeDelta(60 * i, format='sec') for i in range(data.shape[0])])
    cube.extra_coords.add('time', 0, timestamps)
    return cube


@pytest.fixture
def ndcube_3d_rotated(wcs_3d_ln_lt_t_rotated, simple_extra_coords_3d):
    data_rotated = np.array([[[1, 2, 3, 4, 6], [2, 4, 5, 3, 1], [0, -1, 2, 4, 2], [3, 5, 1, 2, 0]],
                             [[2, 4, 5, 1, 3], [1, 5, 2, 2, 4], [2, 3, 4, 0, 5], [0, 1, 2, 3, 4]]])
    mask_rotated = data_rotated >= 0
    cube = NDCube(
        data_rotated,
        wcs_3d_ln_lt_t_rotated,
        mask=mask_rotated,
        uncertainty=data_rotated,
    )
    cube._extra_coords = simple_extra_coords_3d
    return cube


@pytest.fixture
def ndcube_3d_coupled(wcs_3d_ln_lt_l_coupled):
    shape = (128, 256, 512)
    wcs_3d_ln_lt_l_coupled.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    return NDCube(
        data,
        wcs_3d_ln_lt_l_coupled,
        mask=mask,
        uncertainty=data,
    )


@pytest.fixture
def ndcube_3d_coupled_time(wcs_3d_ln_lt_t_coupled):
    shape = (128, 256, 512)
    wcs_3d_ln_lt_t_coupled.array_shape = shape
    data = data_nd(shape)
    return NDCube(
        data,
        wcs_3d_ln_lt_t_coupled,
    )


@pytest.fixture
def ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l):
    return gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l,
                                        1,
                                        Time('2000-01-01', format='fits', scale='utc'))


@pytest.fixture
def ndcube_2d_ln_lt(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln)


@pytest.fixture
def ndcube_2d_ln_lt_uncert(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape)
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty)


@pytest.fixture
def ndcube_2d_ln_lt_mask_uncert(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape)
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    mask = np.zeros(shape, dtype=bool)
    mask[1, 1] = True
    mask[2, 0] = True
    mask[3, 3] = True
    mask[4:6, :4] = True
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, mask=mask)


@pytest.fixture
def ndcube_2d_ln_lt_uncert_ec(wcs_2d_lt_ln):
    shape = (4, 9)
    data_cube = data_nd(shape)
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    cube = NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty)
    cube.extra_coords.add(
        "time", 0,
        Time("2000-01-01 00:00", scale="utc") + TimeDelta(np.arange(shape[0])*60, format="sec"))
    return cube


@pytest.fixture
def ndcube_2d_ln_lt_units(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape).astype(float)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, unit=u.ct)


@pytest.fixture
def ndcube_2d_dask(wcs_2d_lt_ln):
    shape = (8, 4)
    chunks = 2
    data = data_nd(shape).astype(float)
    da = dask.array.asarray(data, chunks=chunks)
    mask = np.zeros(shape, dtype=bool)
    da_mask = dask.array.asarray(mask, chunks=chunks)
    uncert = data * 0.1
    da_uncert = StdDevUncertainty(dask.array.asarray(uncert, chunks=chunks))
    return NDCube(da, wcs=wcs_2d_lt_ln, uncertainty=da_uncert, mask=da_mask, unit=u.J)


@pytest.fixture
def ndcube_2d(request):
    """
    This is a meta fixture for parametrizing all the 2D ndcubes.
    """
    return request.getfixturevalue("ndcube_2d_" + request.param)


@pytest.fixture
def ndcube_1d_l(wcs_1d_l):
    shape = (10,)
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=wcs_1d_l,
                  uncertainty=StdDevUncertainty(data_cube*0.1), unit=u.J)


@pytest.fixture(params=[
    "ndcube_4d_ln_lt_l_t",
    "ndcube_4d_uncertainty",
    "ndcube_4d_mask",
    "ndcube_4d_extra_coords",
    "ndcube_4d_unit_uncertainty",
    "ndcube_3d_ln_lt_l",
    "ndcube_3d_rotated",
    "ndcube_2d_ln_lt",
    "ndcube_2d_ln_lt_units",
    "ndcube_2d_dask",
    "ndcube_1d_l",
])
def all_ndcubes(request):
    """
    All the above ndcube fixtures in order.
    """
    return request.getfixturevalue(request.param)


@pytest.fixture
def ndc(request):
    """
    A fixture for use with indirect to lookup other fixtures.
    """
    return request.getfixturevalue(request.param)


################################################################################
# NDCubeSequence Fixtures
################################################################################


@pytest.fixture
def ndcubesequence_4c_ln_lt_l(ndcube_3d_ln_lt_l):
    cube1 = ndcube_3d_ln_lt_l
    cube2 = ndcube_3d_ln_lt_l
    cube3 = ndcube_3d_ln_lt_l
    cube4 = ndcube_3d_ln_lt_l
    cube2.data[:] *= 2
    cube3.data[:] *= 3
    cube4.data[:] *= 4
    return NDCubeSequence([cube1, cube2, cube3, cube4])


@pytest.fixture
def ndcubesequence_4c_ln_lt_l_cax1(ndcube_3d_ln_lt_l):
    cube1 = ndcube_3d_ln_lt_l
    cube2 = ndcube_3d_ln_lt_l
    cube3 = ndcube_3d_ln_lt_l
    cube4 = ndcube_3d_ln_lt_l
    cube2.data[:] *= 2
    cube3.data[:] *= 3
    cube4.data[:] *= 4
    meta = helpers.ndmeta_et0_pr02((4, 2, 3, 4))
    return NDCubeSequence([cube1, cube2, cube3, cube4], common_axis=1, meta=meta)


@pytest.fixture
def ndcubesequence_3c_l_ln_lt_cax1(wcs_3d_lt_ln_l):
    common_axis = 1

    base_time1 = Time('2000-01-01', format='fits', scale='utc')
    gc1 = GlobalCoords()
    gc1.add('distance', 'custom:distance', 1*u.m)
    cube1 = gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l, 1, base_time1, gc1)

    shape = cube1.data.shape
    base_time2 = base_time1 + TimeDelta([shape[common_axis] * 60], format='sec')
    gc2 = GlobalCoords()
    gc2.add('distance', 'custom:distance', 2*u.m)
    gc2.add('global coord', 'custom:physical_type', 0*u.pix)
    cube2 = gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l, 1, base_time2, gc2)
    cube2.data[:] *= 2

    base_time3 = base_time2 + TimeDelta([shape[common_axis] * 60], format='sec')
    gc3 = GlobalCoords()
    gc3.add('distance', 'custom:distance', 3*u.m)
    cube3 = gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l, 1, base_time3, gc3)
    cube3.data[:] *= 3

    return NDCubeSequence([cube1, cube2, cube3], common_axis=common_axis)


def pytest_runtest_teardown(item):
    # Clear the pyplot figure stack if it is not empty after the test
    # You can see these log messages by passing "-o log_cli=true" to pytest on the command line
    if HAVE_MATPLOTLIB and plt.get_fignums():
        console_logger.info(f"Removing {len(plt.get_fignums())} pyplot figure(s) "
                            f"left open by {item.name}")
        plt.close('all')



import warnings

import matplotlib.pyplot as plt
import numpy as np

import astropy.units as u
from astropy.utils.exceptions import AstropyUserWarning
from astropy.visualization.wcsaxes import WCSAxes

from ndcube.utils.exceptions import warn_user
from . import plotting_utils as utils
from .base import BasePlotter
from .descriptor import MISSING_ANIMATORS_ERROR_MSG

__all__ = ['MatplotlibPlotter']


class MatplotlibPlotter(BasePlotter):
    """
    Provide visualization methods for NDCube which use `matplotlib`.
    """

    def plot(self, axes=None, plot_axes=None, axes_coordinates=None,
             axes_units=None, data_unit=None, wcs=None, **kwargs):
        """
        Visualize the `~ndcube.NDCube`.

        Parameters
        ----------
        axes: `~astropy.visualization.wcsaxes.WCSAxes` or None:, optional
            The axes to plot onto. If None the current axes will be used.

        plot_axes: `list`, optional
            A list of length equal to the number of pixel dimensions in array axis order.
            This list selects which cube axes are displayed on which plot axes.
            For an image plot this list should contain ``'x'`` and ``'y'`` for the
            plot axes and `None` for all the other elements. For a line plot it
            should only contain ``'x'`` and `None` for all the other elements.

        axes_unit: `list`, optional
            A list of length equal to the number of world dimensions specifying
            the units of each axis, or `None` to use the default unit for that
            axis.

        axes_coordinates: `list`, optional
            A list of length equal to the number of pixel dimensions. For each
            axis the value of the list should either be a string giving the
            world axis type or `None` to use the default axis from the WCS.

        data_unit: `astropy.units.Unit`
            The data is changed to the unit given or the ``NDCube.unit`` if not
            given.

        wcs: `astropy.wcs.wcsapi.BaseHighLevelWCS`
            The WCS object to define the coordinates of the plot axes.

        kwargs :
            Additional keyword arguments are given to the underlying plotting infrastructure
            which depends on the dimensionality of the data and whether 1 or 2 plot_axes are
            defined:
            - Animations: `mpl_animators.ArrayAnimatorWCS`
            - Static 2-D images: `matplotlib.pyplot.imshow`
            - Static 1-D line plots: `matplotlib.pyplot.plot`
        """
        naxis = self._ndcube.wcs.pixel_n_dim

        if not axes_coordinates:
            axes_coordinates = [...]
            plot_wcs = self._ndcube.wcs.low_level_wcs
        else:
            plot_wcs = self._ndcube.combined_wcs.low_level_wcs
        if wcs is not None:
            plot_wcs = wcs.low_level_wcs

        # Check kwargs are in consistent formats and set default values if not done so by user.
        plot_axes, axes_coordinates, axes_units = utils.prep_plot_kwargs(
            len(self._ndcube.shape), plot_wcs, plot_axes, axes_coordinates, axes_units)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore', AstropyUserWarning)
            if naxis == 1:
                ax = self._plot_1D_cube(plot_wcs, axes, axes_coordinates,
                                        axes_units, data_unit, **kwargs)

            elif naxis == 2 and 'y' in plot_axes:
                ax = self._plot_2D_cube(plot_wcs, axes, plot_axes, axes_coordinates,
                                        axes_units, data_unit, **kwargs)
            else:
                ax = self._animate_cube(plot_wcs, plot_axes=plot_axes,
                                        axes_coordinates=axes_coordinates,
                                        axes_units=axes_units, data_unit=data_unit, **kwargs)

        return ax

    def _not_visible_coords(self, axes, axes_coordinates):
        """
        Based on an axes object and axes_coords, work out which coords should not be visible.
        """
        visible_coords = {item[1] for item in axes.coords._aliases.items() if item[0] in axes_coordinates}
        return set(axes.coords._aliases.values()).difference(visible_coords)

    def _apply_axes_coordinates(self, axes, axes_coordinates):
        """
        Hide ticks and labels for non-visible axes based on axes_coordinates.
        """
        for coord_index in self._not_visible_coords(axes, axes_coordinates):
            axes.coords[coord_index].set_ticks_visible(False)
            axes.coords[coord_index].set_ticklabel_visible(False)

    def _plot_1D_cube(self, wcs, axes=None, axes_coordinates=None, axes_units=None,
                      data_unit=None, **kwargs):
        if axes is None:
            axes = plt.subplot(projection=wcs)

        self._apply_axes_coordinates(axes, axes_coordinates)

        default_ylabel = "Data"

        # Derive y-axis coordinates, uncertainty and unit from the NDCube's data.
        yerror = self._ndcube.uncertainty.array if (self._ndcube.uncertainty is not None) else None
        ydata = self._ndcube.data

        if self._ndcube.unit is None:
            if data_unit is not None:
                raise TypeError("Can only set y-axis unit if self._ndcube.unit is set to a "
                                "compatible unit.")
        else:
            if data_unit is not None:
                ydata = u.Quantity(ydata, unit=self._ndcube.unit).to_value(data_unit)
                if yerror is not None:
                    yerror = u.Quantity(yerror, self._ndcube.unit).to_value(data_unit)
            else:
                data_unit = self._ndcube.unit

            default_ylabel += f" [{data_unit}]"

        # Combine data and uncertainty with mask.
        if self._ndcube.mask is not None:
            ydata = np.ma.masked_array(ydata, self._ndcube.mask)

            if yerror is not None:
                yerror = np.ma.masked_array(yerror, self._ndcube.mask)

        if yerror is not None:
            # We plot against pixel coordinates
            axes.errorbar(np.arange(len(ydata)), ydata, yerr=yerror, **kwargs)
        else:
            axes.plot(ydata, **kwargs)

        axes.set_ylabel(default_ylabel)

        utils.set_wcsaxes_format_units(axes.coords, wcs, axes_units)

        return axes

    def _plot_2D_cube(self, wcs, axes=None, plot_axes=None, axes_coordinates=None,
                      axes_units=None, data_unit=None, **kwargs):
        if axes is None:
            axes = plt.subplot(projection=wcs, slices=plot_axes)

        if axes and plot_axes:
            axes.reset_wcs(wcs=wcs, slices=plot_axes)

        utils.set_wcsaxes_format_units(axes.coords, wcs, axes_units)

        self._apply_axes_coordinates(axes, axes_coordinates)

        data = self._ndcube.data
        if data_unit is not None:
            # If user set data_unit, convert dat to desired unit if self._ndcube.unit set.
            if self._ndcube.unit is None:
                raise TypeError("Can only set data_unit if NDCube.unit is set.")
            data = u.Quantity(self._ndcube.data, unit=self._ndcube.unit).to_value(data_unit)

        if self._ndcube.mask is not None:
            data = np.ma.masked_array(data, self._ndcube.mask)

        if plot_axes.index('x') > plot_axes.index('y'):
            data = data.T

        # Plot data
        im = axes.imshow(data, **kwargs)

        # Set current axes/image if pyplot is being used (makes colorbar work)
        for i in plt.get_fignums():
            if axes in plt.figure(i).axes:
                plt.sca(axes)
                plt.sci(im)

        return axes

    def _animate_cube(self, wcs, plot_axes=None, axes_coordinates=None,
                      axes_units=None, data_unit=None, **kwargs):
        try:
            from mpl_animators import ArrayAnimatorWCS
        except ImportError as e:
            raise ImportError(MISSING_ANIMATORS_ERROR_MSG) from e

        # Derive inputs for animation object and instantiate.
        data, wcs, plot_axes, coord_params = self._prep_animate_args(wcs, plot_axes,
                                                                     axes_units, data_unit)

        ax = ArrayAnimatorWCS(data, wcs, plot_axes, coord_params=coord_params, **kwargs)

        # We need to modify the visible axes after the axes object has been created.
        # This call affects only the initial draw
        self._apply_axes_coordinates(ax.axes, axes_coordinates)

        # This changes the parameters for future iterations
        for hidden in self._not_visible_coords(ax.axes, axes_coordinates):
            if hidden in ax.coord_params:
                param = ax.coord_params[hidden]
            else:
                param = {}

            param['ticks'] = False
            ax.coord_params[hidden] = param

        return ax

    def _as_mpl_axes(self):
        """
        Compatibility hook for Matplotlib and WCSAxes.
        This functionality requires the WCSAxes package to work. The reason
        we include this here is that it allows users to use WCSAxes without
        having to explicitly import WCSAxes
        With this method, one can do::
            fig = plt.figure()  # doctest: +SKIP
            ax = plt.subplot(projection=my_ndcube)  # doctest: +SKIP
        and this will generate a plot with the correct WCS coordinates on the
        axes. See https://wcsaxes.readthedocs.io for more information.
        """
        kwargs = {'wcs': self._ndcube.wcs}
        n_dim = len(self._ndcube.shape)
        if n_dim > 2:
            kwargs['slices'] = ['x', 'y'] + [None] * (n_dim - 2)
        return WCSAxes, kwargs

    def _prep_animate_args(self, wcs, plot_axes, axes_units, data_unit):
        # If data_unit set, convert data to that unit
        if data_unit is None:
            data = self._ndcube.data
        else:
            data = u.Quantity(self._ndcube.data, unit=self._ndcube.unit, copy=False).to_value(data_unit)

        # Combine data values with mask.
        if self._ndcube.mask is not None:
            data = np.ma.masked_array(data, self._ndcube.mask)

        coord_params = {}
        if axes_units is not None:
            for axis_unit, coord_name in zip(axes_units, wcs.world_axis_physical_types):
                coord_params[coord_name] = {'format_unit': axis_unit}

        # TODO: Add support for transposing the array.
        if 'y' in plot_axes and plot_axes.index('y') < plot_axes.index('x'):
            warn_user(
                "Animating a NDCube does not support transposing the array. The world axes "
                "may not display as expected because the array will not be transposed."
            )
        plot_axes = [p if p is not None else 0 for p in plot_axes]

        return data, wcs, plot_axes, coord_params




import warnings

import numpy as np
import pytest

import astropy.units as u
from astropy.tests.helper import assert_quantity_allclose
from astropy.time import Time, TimeDelta

from ndcube import NDCube, NDCubeSequence, NDMeta
from ndcube.tests import helpers


def derive_sliced_cube_dims(orig_cube_dims, tuple_item):
    expected_cube_dims = list(orig_cube_dims)
    len_cube_item = len(tuple_item) - 1
    if len_cube_item > 0:
        cube_item = tuple_item[1:]
        for i, s in zip(np.arange(len_cube_item)[::-1], cube_item[::-1]):
            if isinstance(s, int):
                del expected_cube_dims[i]
            else:
                expected_cube_dims[i] = float(s.stop - s.start)
    return tuple(expected_cube_dims)


@pytest.mark.parametrize(("ndc", "item"),
                         [
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0:1], ),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0:1, 0:2]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0:1, 1]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[1:3, 1, 0:2])
                         ],
                         indirect=('ndc',))
def test_slice_sequence_axis(ndc, item):
    # Calculate expected dimensions of cubes with sequence after slicing.
    tuple_item = item if isinstance(item, tuple) else (item,)
    expected_cube0_dims = derive_sliced_cube_dims(ndc.data[tuple_item[0]][0].shape, tuple_item)
    # Assert output is as expected.
    sliced_sequence = ndc[item]
    assert isinstance(sliced_sequence, NDCubeSequence)
    assert int(sliced_sequence.shape[0]) == tuple_item[0].stop - tuple_item[0].start
    assert np.all(sliced_sequence[0].shape == expected_cube0_dims)


@pytest.mark.parametrize(("ndc", "item"),
                         [
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[1, 0:1]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[2, 1]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[3, 1, 0:2])
                         ],
                         indirect=("ndc",))
def test_extract_ndcube(ndc, item):
    cube = ndc[item]
    tuple_item = item if isinstance(item, tuple) else (item,)
    expected_cube_dims = derive_sliced_cube_dims(ndc.data[tuple_item[0]].shape, tuple_item)
    assert isinstance(cube, NDCube)
    assert np.all(cube.shape == expected_cube_dims)


@pytest.mark.parametrize(("ndc", "item", "expected_common_axis"),
                         [
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, 0], 0),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, 0:1, 0:2], 1),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, :, :, 1], 1),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, :, 0], None)
                         ],
                         indirect=("ndc",))
def test_slice_common_axis(ndc, item, expected_common_axis):
    sliced_sequence = ndc[item]
    assert sliced_sequence._common_axis == expected_common_axis


@pytest.mark.parametrize(("ndc", "item", "expected_shape"),
                         [
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, 1:7], (3, 2, (2, 3, 1), 4)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0, 1:7], (3, (2, 3, 1), 4)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, 2:4], (2, 2, 1, 4)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, 0:6], (2, 2, 3, 4)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0, 0:6], (2, 3, 4)),
                         ],
                         indirect=("ndc",))
def test_index_as_cube(ndc, item, expected_shape):
    assert (ndc.index_as_cube[item].shape == expected_shape)


@pytest.mark.parametrize(("ndc", "axis", "expected_shape"),
                         [
                             ("ndcubesequence_4c_ln_lt_l", 0, (8,
                                                               3,
                                                               4)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", 1, (12,
                                                                    2,
                                                                    4))
                         ],
                         indirect=("ndc",))
def test_explode_along_axis_common_axis_none(ndc, axis, expected_shape):
    exploded_sequence = ndc.explode_along_axis(axis)
    assert np.all(exploded_sequence.shape == expected_shape)
    assert exploded_sequence._common_axis is None


@pytest.mark.parametrize("ndc", (['ndcubesequence_4c_ln_lt_l_cax1']), indirect=("ndc",))
def test_explode_along_axis_common_axis_same(ndc):
    exploded_sequence = ndc.explode_along_axis(2)
    assert exploded_sequence.shape == (16, 2, 3)
    assert exploded_sequence._common_axis == ndc._common_axis


@pytest.mark.parametrize("ndc", (['ndcubesequence_4c_ln_lt_l_cax1']), indirect=("ndc",))
def test_explode_along_axis_common_axis_changed(ndc):
    exploded_sequence = ndc.explode_along_axis(0)
    assert exploded_sequence.shape == (8, 3, 4)
    assert exploded_sequence._common_axis == ndc._common_axis - 1


@pytest.mark.parametrize(("ndc", "expected_shape"),
                         [
                             ("ndcubesequence_4c_ln_lt_l_cax1", (4,
                                                                 2.,
                                                                 3.,
                                                                 4.)),
                         ],
                         indirect=("ndc",))
def test_shape(ndc, expected_shape):
    np.testing.assert_array_equal(ndc.shape, expected_shape)


@pytest.mark.parametrize(("ndc", "expected_shape"),
                         [
                             ("ndcubesequence_4c_ln_lt_l_cax1", [2., 12, 4]),
                         ],
                         indirect=("ndc",))
def test_cube_like_shape(ndc, expected_shape):
    assert np.all(ndc.cube_like_shape == expected_shape)


@pytest.mark.parametrize(("ndc", "expected_dimensions"),
                         [
                             ("ndcubesequence_4c_ln_lt_l_cax1", tuple(u.Quantity(d, unit=u.pix) for d in [2. , 12, 4])),
                         ],
                         indirect=("ndc",))
def test_cube_like_dimensions(ndc, expected_dimensions):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for ndc_dim, exp_dim in zip(ndc.cube_like_dimensions, expected_dimensions):
            assert_quantity_allclose(ndc_dim, exp_dim)


@pytest.mark.parametrize("ndc", (["ndcubesequence_4c_ln_lt_l"]), indirect=("ndc",))
def test_cube_like_shape_error(ndc):
    with pytest.raises(TypeError):
        ndc.cube_like_shape


@pytest.mark.parametrize("ndc", (["ndcubesequence_3c_l_ln_lt_cax1"]), indirect=("ndc",))
def test_common_axis_coords(ndc):
    # Construct expected skycoord
    common_coords = [cube.axis_world_coords('lon') for cube in ndc]
    expected_skycoords = []
    for cube_coords in common_coords:
        expected_skycoords += [cube_coords[0][i] for i in range(len(cube_coords[0]))]
    # Construct expected Times
    base_time = Time('2000-01-01', format='fits', scale='utc')
    expected_times = [base_time + TimeDelta(60*i, format='sec') for i in range(15)]
    # Run test function.
    output = ndc.common_axis_coords
    #  Check right number of coords returned.
    assert len(output) == 2
    output_skycoords, output_times = output
    # Check SkyCoords are equal.
    for output_coord, expected_coord in zip(output_skycoords, expected_skycoords):
        assert all(output_coord == expected_coord)
    # Check times are equal
    for output_time, expected_time in zip(output_times, expected_times):
        td = output_time - expected_time
        assert u.allclose(td.to(u.s), 0*u.s, atol=1e-10*u.s)


@pytest.mark.parametrize("ndc", (["ndcubesequence_3c_l_ln_lt_cax1"]), indirect=("ndc",))
def test_sequence_axis_coords(ndc):
    expected = {'distance': [1*u.m, 2*u.m, 3*u.m]}
    output = ndc.sequence_axis_coords
    assert output == expected


def test_crop(ndcubesequence_4c_ln_lt_l):
    seq = ndcubesequence_4c_ln_lt_l
    intervals = seq[0].wcs.array_index_to_world([1, 2], [0, 1], [0, 2])
    lower_corner = [coord[0] for coord in intervals]
    upper_corner = [coord[-1] for coord in intervals]
    expected = seq[:, 1:3, 0:2, 0:3]
    output = seq.crop(lower_corner, upper_corner)
    helpers.assert_cubesequences_equal(output, expected)


def test_crop_by_values(ndcubesequence_4c_ln_lt_l):
    seq = ndcubesequence_4c_ln_lt_l
    intervals = seq[0].wcs.array_index_to_world_values([1, 2], [0, 1], [0, 2])
    units = [u.m, u.deg, u.deg]
    lower_corner = [coord[0] * unit for coord, unit in zip(intervals, units)]
    upper_corner = [coord[-1] * unit for coord, unit in zip(intervals, units)]
    # Ensure some quantities are in units different from each other
    # and those stored in the WCS.
    lower_corner[0] = lower_corner[0].to(units[0])
    lower_corner[-1] = lower_corner[-1].to(units[-1])
    upper_corner[-1] = upper_corner[-1].to(units[-1])
    expected = seq[:, 1:3, 0:2, 0:3]
    output = seq.crop_by_values(lower_corner, upper_corner)
    helpers.assert_cubesequences_equal(output, expected)


def test_slice_meta(ndcubesequence_4c_ln_lt_l_cax1):
    seq = ndcubesequence_4c_ln_lt_l_cax1
    sliced_seq = seq[:, :, 0]
    expected_meta = NDMeta({"salutation": "hello",
                            "exposure time": u.Quantity([2] * 4, unit=u.s),
                            "pixel response": u.Quantity([100] * 4, unit=u.percent)},
                           axes={"exposure time": 0, "pixel response": 0}, data_shape=(4, 2, 4))
    helpers.assert_metas_equal(sliced_seq.meta, expected_meta)



import numbers

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs.wcsapi import HighLevelWCSWrapper

from ndcube.wcs.wrappers import ResampledLowLevelWCS


@pytest.fixture
def celestial_wcs(request):
    return request.getfixturevalue(request.param)


EXPECTED_2D_REPR_NUMPY2 = """
ResampledLowLevelWCS Transformation

This transformation has 2 pixel and 2 world dimensions

Array shape (Numpy order): (2, 15)

Pixel Dim  Axis Name  Data size  Bounds
        0  None              15  (-2.5, 12.5)
        1  None         2.33333  (0.3333333333333333, 2.3333333333333335)

World Dim  Axis Name        Physical Type  Units
        0  Right Ascension  pos.eq.ra      deg
        1  Declination      pos.eq.dec     deg

Correlation between pixel and world axes:

           Pixel Dim
World Dim    0    1
        0  yes  yes
        1  yes  yes
""".strip()
EXPECTED_2D_REPR_NUMPY1 = """
ResampledLowLevelWCS Transformation

This transformation has 2 pixel and 2 world dimensions

Array shape (Numpy order): (2.3333333333333335, 15.0)

Pixel Dim  Axis Name  Data size  Bounds
        0  None              15  (-2.5, 12.5)
        1  None         2.33333  (0.3333333333333333, 2.3333333333333335)

World Dim  Axis Name        Physical Type  Units
        0  Right Ascension  pos.eq.ra      deg
        1  Declination      pos.eq.dec     deg

Correlation between pixel and world axes:

           Pixel Dim
World Dim    0    1
        0  yes  yes
        1  yes  yes
""".strip()

@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs', 'celestial_2d_fitswcs'],
                         indirect=True)
def test_2d(celestial_wcs):

    # Upsample along the first pixel dimension and downsample along the second
    # pixel dimension.
    wcs = ResampledLowLevelWCS(celestial_wcs, [0.4, 3])

    # The following shouldn't change compared to the original WCS
    assert wcs.pixel_n_dim == 2
    assert wcs.world_n_dim == 2
    assert tuple(wcs.world_axis_physical_types) == ('pos.eq.ra', 'pos.eq.dec')
    assert tuple(wcs.world_axis_units) == ('deg', 'deg')
    assert tuple(wcs.pixel_axis_names) == ('', '')
    assert tuple(wcs.world_axis_names) == ('Right Ascension',
                                           'Declination')
    assert_equal(wcs.axis_correlation_matrix, np.ones((2, 2)))

    # Shapes and bounds should be floating-point if needed
    assert_allclose(wcs.pixel_shape, (15, 7/3))
    assert_allclose(wcs.array_shape, (7/3, 15))
    assert_allclose(wcs.pixel_bounds, ((-2.5, 12.5), (1/3, 7/3)))

    pixel_scalar = (2.3, 1.2)
    world_scalar = (12.16, -4.8)
    assert_allclose(wcs.pixel_to_world_values(*pixel_scalar), world_scalar)
    assert_allclose(wcs.array_index_to_world_values(*pixel_scalar[::-1]), world_scalar)
    assert_allclose(wcs.world_to_pixel_values(*world_scalar), pixel_scalar)
    assert_allclose(wcs.world_to_array_index_values(*world_scalar), [1, 2])

    EXPECTED_2D_REPR = EXPECTED_2D_REPR_NUMPY2 if np.__version__ >= '2.0.0' else EXPECTED_2D_REPR_NUMPY1
    assert str(wcs) == EXPECTED_2D_REPR
    assert EXPECTED_2D_REPR in repr(wcs)

    celestial_wcs.pixel_bounds = None
    pixel_array = (np.array([2.3, 2.4]),
                   np.array([4.3, 4.4]))
    world_array = (np.array([12.16, 12.08]),
                   np.array([13.8, 14.4]))
    assert_allclose(wcs.pixel_to_world_values(*pixel_array), world_array)
    assert_allclose(wcs.array_index_to_world_values(*pixel_array[::-1]), world_array)
    assert_allclose(wcs.world_to_pixel_values(*world_array), pixel_array)
    assert_allclose(wcs.world_to_array_index_values(*world_array),
                    [[4, 4], [2, 2]])

    wcs_hl = HighLevelWCSWrapper(wcs)

    celestial = wcs_hl.pixel_to_world(*pixel_scalar)
    assert isinstance(celestial, SkyCoord)
    assert_quantity_allclose(celestial.ra, world_scalar[0] * u.deg)
    assert_quantity_allclose(celestial.dec, world_scalar[1] * u.deg)

    celestial = wcs_hl.pixel_to_world(*pixel_array)
    assert isinstance(celestial, SkyCoord)
    assert_quantity_allclose(celestial.ra, world_array[0] * u.deg)
    assert_quantity_allclose(celestial.dec, world_array[1] * u.deg)


@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs', 'celestial_2d_fitswcs'],
                         indirect=True)
def test_scalar_factor(celestial_wcs):

    celestial_wcs.pixel_bounds = None
    wcs = ResampledLowLevelWCS(celestial_wcs, 2)

    pixel_scalar = (2.3, 4.3)
    world_scalar = (4.8, 5.2)
    assert_allclose(wcs.pixel_to_world_values(*pixel_scalar), world_scalar)
    assert_allclose(wcs.array_index_to_world_values(*pixel_scalar[::-1]), world_scalar)
    assert_allclose(wcs.world_to_pixel_values(*world_scalar), pixel_scalar)
    assert_allclose(wcs.world_to_array_index_values(*world_scalar), [4, 2])


@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs', 'celestial_2d_fitswcs'],
                         indirect=True)
def test_offset(celestial_wcs):
    celestial_wcs.pixel_bounds = None
    offset = 1
    factor = 2
    wcs = ResampledLowLevelWCS(celestial_wcs, factor, offset=offset)

    pixel_scalar = (2.3, 4.3)
    world_scalar = celestial_wcs.pixel_to_world_values(*[p * factor + offset
                                                         for p in pixel_scalar])

    assert_allclose(wcs.pixel_to_world_values(*pixel_scalar), world_scalar)
    assert_allclose(wcs.array_index_to_world_values(*pixel_scalar[::-1]), world_scalar)
    assert_allclose(wcs.world_to_pixel_values(*world_scalar), pixel_scalar)
    assert_allclose(wcs.world_to_array_index_values(*world_scalar), [4, 2])


@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs'],
                         indirect=True)
def test_factor_wrong_length_error(celestial_wcs):
    with pytest.raises(ValueError):
        ResampledLowLevelWCS(celestial_wcs, [2] * 3)


@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs'],
                         indirect=True)
def test_scalar_wrong_length_error(celestial_wcs):
    with pytest.raises(ValueError):
        ResampledLowLevelWCS(celestial_wcs, 2, offset=[1] * 3)


@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs', 'celestial_2d_fitswcs'],
                         indirect=True)
def test_int_fraction_pixel_shape(celestial_wcs):
    # Some fractional factors are not representable by exact floats, e.g. 1/3.
    # However, it is still desirable for the pixel shape to return ints in these cases.
    # This test checks that this is the case.
    wcs = ResampledLowLevelWCS(celestial_wcs, 1/3)
    assert wcs.pixel_shape == (18, 21)
    for dim in wcs.pixel_shape:
        assert isinstance(dim, numbers.Integral)



import copy

import numpy as np
import pytest

from ndcube.meta import NDMeta
from .helpers import assert_metas_equal


@pytest.fixture
def basic_meta_values():
    return {"a": "hello",
            "b": list(range(10, 25, 10)),
            "c": np.array([[1, 2, 3, 4], [10, 20, 30, 40], [100, 200, 300, 400]]),
            "d": list(range(3, 13, 3)),
            "e": list(range(2, 8, 2)),
            "f": "world",
            "g": ["hello", "world", "!"]
            }


@pytest.fixture
def basic_key_comments():
    return {"a": "Comment A",
            "b": "Comment B",
            "c": "Comment C",
            }


@pytest.fixture
def basic_axes():
    return {"b": 0,
            "c": (1, 3),
            "d": (3,),
            "e": 1,
            "f": 0,
            "g": (0, 1, 3)
            }


@pytest.fixture
def basic_data_shape():
    return (2, 3, 0, 4)

@pytest.fixture
def basic_meta(basic_meta_values, basic_key_comments, basic_axes):
    return NDMeta(basic_meta_values, basic_key_comments, basic_axes)


@pytest.fixture
def no_shape_meta():
    return NDMeta({"a": "hello"})


def test_shape(basic_meta, basic_data_shape):
    meta = basic_meta
    shape = np.asarray(basic_data_shape)
    assert all(meta.data_shape == shape)


def test_slice_axis_with_no_meta(basic_meta):
    meta = basic_meta
    output = meta.slice[:, :, 0]
    expected = copy.deepcopy(meta)
    expected._data_shape = meta._data_shape[[0, 1, 3]]
    expected._axes["c"] = (1, 2)
    expected._axes["d"] = (2,)
    expected._axes["g"] = (0, 1, 2)
    assert_metas_equal(output, expected)


def test_slice_away_independent_axis(basic_meta):
    meta = basic_meta
    item = 0
    output = meta.slice[item]
    values = dict(meta.items())
    values["b"] = values["b"][0]
    values["g"] = ["world", "!"]
    del values["f"]
    key_comments = meta.key_comments
    axes = copy.deepcopy(meta.axes)
    del axes["b"]
    del axes["f"]
    axes["c"] -= 1
    axes["d"] -= 1
    axes["e"] -= 1
    axes["g"] = (0, 2)
    expected = NDMeta(values, key_comments, axes)
    assert_metas_equal(output, expected)


def test_slice_away_independent_and_dependent_axis(basic_meta):
    meta = basic_meta
    item = (0, 1)
    output = meta.slice[item]
    values = dict(meta.items())
    del values["f"]
    values["b"] = values["b"][0]
    values["c"] = values["c"][1]
    values["e"] = values["e"][1]
    values["g"] = "!"
    key_comments = meta.key_comments
    axes = copy.deepcopy(meta.axes)
    del axes["b"]
    del axes["e"]
    del axes["f"]
    axes["c"] = 1
    axes["d"] = 1
    axes["g"] = 1
    expected = NDMeta(values, key_comments, axes)
    assert_metas_equal(output, expected)


def test_slice_dependent_axes(basic_meta):
    meta = basic_meta
    output = meta.slice[:, 1:3, :, 1]
    values = dict(meta.items())
    values["c"] = values["c"][1:3, 1]
    values["d"] = values["d"][1]
    values["e"] = values["e"][1:3]
    values["g"] = values["g"][:2]
    key_comments = meta.key_comments
    axes = copy.deepcopy(meta.axes)
    del axes["d"]
    axes["c"] = 1
    axes["g"] = (0, 1)
    expected = NDMeta(values, key_comments, axes)
    expected._data_shape = np.array([2, 2, 0])
    assert_metas_equal(output, expected)


def test_slice_by_str(basic_meta):
    meta = basic_meta
    assert meta["a"] == "hello"
    assert meta["b"] == list(range(10, 25, 10))


def test_add1(basic_meta):
    meta = basic_meta
    name = "z"
    value = 100
    key_comment = "Comment E"
    meta.add(name, value, key_comment, None)
    assert name in meta.keys()
    assert meta[name] == value
    assert meta.key_comments[name] == key_comment
    assert meta.axes.get(name, None) is None


def test_add2(basic_meta):
    meta = basic_meta
    name = "z"
    value = list(range(2))
    axis = 0
    meta.add(name, value, None, axis)
    assert name in meta.keys()
    assert meta[name] == value
    assert meta.key_comments.get(name, None) is None
    assert meta.axes[name] == np.array([axis])


def test_add3():
    meta = NDMeta(data_shape=[5, 3, 2])
    name = "axis name"
    value = np.array(["a", "b", "c", "d"])
    axis = (0, 1, 2, 3)
    meta.add(name, value, axes=axis)
    assert name in meta.keys()
    assert (meta[name] == value).all()
    assert (meta.axes[name] == np.array([axis])).all()
    assert (meta.data_shape == np.array([5, 3, 2, 0])).all()


def test_add_overwrite(basic_meta):
    meta = basic_meta
    name = "a"
    value = "goodbye"
    meta.add(name, value, None, None, overwrite=True)
    assert meta[name] == value


def test_add_overwrite_error(basic_meta):
    meta = basic_meta
    with pytest.raises(KeyError):
        meta.add("a", "world", None, None)


def test_del(basic_meta):
    meta = basic_meta
    name = "b"
    del meta[name]
    assert name not in meta.keys()
    assert name not in meta.key_comments.keys()
    assert name not in meta.axes.keys()


def test_rebin(basic_meta):
    meta = basic_meta
    bin_shape = (2, 1, 1, 2)
    output = meta.rebin(bin_shape)
    # Build expected result.
    expected = copy.deepcopy(meta)
    del expected._axes["b"]
    del expected._axes["c"]
    del expected._axes["d"]
    expected._data_shape = np.array([1, 3, 0, 2], dtype=int)
    assert_metas_equal(output, expected)


def test_data_shape_setter(basic_meta):
    meta = basic_meta
    new_shape = np.array([2, 3, 5, 4], dtype=int)
    meta.data_shape = new_shape
    assert tuple(meta.data_shape) == tuple(new_shape)


def test_data_shape_setter_error(basic_meta):
    meta = basic_meta
    with pytest.raises(ValueError):
        meta.data_shape = (3, 3, 0, 4)



import pytest

import astropy.units as u

import ndcube.visualization.plotting_utils as utils


@pytest.mark.parametrize(("ndim", "plist", "output"), [
    (2, ['x', 'y'], ['x', 'y']),
    (2, [..., 'x', 'y'], ['x', 'y']),
    (2, ['x', 'y', ...], ['x', 'y']),
    (3, ['x', ...], ['x', None, None]),
    (4, ['x', ..., 'y'], ['x', None, None, 'y']),
    (5, [..., 'x'], [None, None, None, None, 'x']),
    (5, [..., 'x', None], [None, None, None, 'x', None]),
    (5, [None, ..., 'x', None, 'y'], [None, None, 'x', None, 'y']),
])
def test_expand_ellipsis(ndim, plist, output):
    result = utils._expand_ellipsis(ndim, plist)
    assert result == output


def test_expand_ellipsis_error():
    with pytest.raises(IndexError):
        utils._expand_ellipsis(1, (..., 'x', ...))


def test_prep_plot_kwargs_errors(ndcube_4d_ln_lt_l_t):
    """
    Check a whole bunch of different error conditions.
    """
    # plot_axes has incorrect length
    with pytest.raises(ValueError):
        utils.prep_plot_kwargs(4, ndcube_4d_ln_lt_l_t.wcs, ['wibble'], None, None)

    # axes_coordinates is not in world_axis_physical_types
    with pytest.raises(ValueError):
        utils.prep_plot_kwargs(4, ndcube_4d_ln_lt_l_t.wcs, None, [..., "wibble"], None)

    # axes_coordinates has incorrect type
    with pytest.raises(TypeError):
        utils.prep_plot_kwargs(4, ndcube_4d_ln_lt_l_t.wcs, None, [..., 10], None)

    # axes_units has incorrect length
    with pytest.raises(ValueError):
        utils.prep_plot_kwargs(4, ndcube_4d_ln_lt_l_t.wcs, None, None, ['m'])

    # axes_units has incorrect type
    with pytest.raises(TypeError):
        utils.prep_plot_kwargs(4, ndcube_4d_ln_lt_l_t.wcs, None, None, [[], ...])

    with pytest.raises(u.UnitsError):
        utils.prep_plot_kwargs(4, ndcube_4d_ln_lt_l_t.wcs, None, None, [u.eV, u.m, u.m, u.m])


@pytest.mark.parametrize(("ndcube_2d", "args", "output"), [
    ("ln_lt",
     (None, None, None),
     (['x', 'y'], None, None)),
    ("ln_lt",
     (None, [..., 'custom:pos.helioprojective.lon'], None),
     (['x', 'y'], ['custom:pos.helioprojective.lat', 'custom:pos.helioprojective.lon'], None)),
    ("ln_lt",
     (None, None, [u.deg, 'arcsec']),
     (['x', 'y'], None, [u.arcsec, u.deg])),
], indirect=['ndcube_2d'])
def test_prep_plot_kwargs(ndcube_2d, args, output):
    result = utils.prep_plot_kwargs(2, ndcube_2d.wcs, *args)
    assert result == output




import pytest

from ndcube.utils import collection as collection_utils


@pytest.mark.parametrize(("data_dimensions1", "data_dimensions2", "data_axes1", "data_axes2"), [
    ([3., 4., 5.], [3., 5., 15.], (0, 2), (0, 1))])
def test_assert_aligned_axes_compatible(data_dimensions1, data_dimensions2,
                                        data_axes1, data_axes2):
    collection_utils.assert_aligned_axes_compatible(data_dimensions1, data_dimensions2,
                                                    data_axes1, data_axes2)


@pytest.mark.parametrize(("data_dimensions1", "data_dimensions2", "data_axes1", "data_axes2"), [
    ([3., 4., 5.], [3., 5., 15.], (0, 1), (0, 1)),
    ([3., 4., 5.], [3., 5., 15.], (0, 1), (0, 1, 2)),
    ([3., 4., 5.], [3., 5., 15.], (0, 1), None)])
def test_assert_aligned_axes_compatible_error(data_dimensions1, data_dimensions2,
                                              data_axes1, data_axes2):
    with pytest.raises(ValueError):
        collection_utils.assert_aligned_axes_compatible(data_dimensions1, data_dimensions2,
                                                        data_axes1, data_axes2)



"""
======
ndcube
======

A package for multi-dimensional contiguous and non-contiguous coordinate-aware arrays.

* Homepage: https://github.com/sunpy/ndcube
* Documentation: https://docs.sunpy.org/projects/ndcube/
"""
from .extra_coords.extra_coords import ExtraCoords, ExtraCoordsABC
from .global_coords import GlobalCoords, GlobalCoordsABC
from .meta import NDMeta
from .ndcollection import NDCollection
from .ndcube import NDCube, NDCubeBase
from .ndcube_sequence import NDCubeSequence, NDCubeSequenceBase
from .version import version as __version__


__all__ = [
    "ExtraCoords",
    "ExtraCoordsABC",
    "GlobalCoords",
    "GlobalCoordsABC",
    "NDCollection",
    'NDCube',
    "NDCubeBase",
    'NDCubeSequence',
    "NDCubeSequenceBase",
    "NDMeta",
    "__version__",
]
